import logging
from typing import List, Optional

from pydantic import BaseModel


class FilterList(BaseModel):
    blacklist: Optional[List[str]] = []
    whitelist: Optional[List[str]] = []


class Filters(BaseModel):
    text: Optional[FilterList]


class TgcfFilter:
    id = "filter"

    def __init__(self, data):
        print("tgcf filter data loaded")
        self.filters = Filters(**data)
        logging.info(self.filters)

    def modify(self, message):
        msg_text = message.text
        if not msg_text:
            return message

        if self.text_safe(msg_text, self.filters.text):
            return message

    def text_safe(self, text, flist: FilterList):
        for forbidden in flist.blacklist:
            if forbidden in text:
                return False
        if not flist.whitelist:
            return True
        for allowed in flist.whitelist:
            if allowed in text:
                return True



# a custom config parser

import logging
import os
import sys
from typing import Dict, List, Optional,Union

import yaml
from pydantic import BaseModel
from telethon.sessions import StringSession

CONFIG_FILE_NAME = "tgcf.config.yml"
CONFIG_ENV_VAR_NAME = "TGCF_CONFIG"


class Forward(BaseModel):
    source: Union[int,str]
    dest: List[Union[int,str]]
    offset: Optional[int] = 0


class Config(BaseModel):
    forwards: List[Forward]
    show_forwarded_from: Optional[bool] = False
    plugins: Optional[Dict] = {}


def detect_config_type() -> int:
    # return 1 when tgcf.config.yml
    # 2 when env var
    # else terminate
    tutorial_link = "Learn more http://bit.ly/configure-tgcf"

    if CONFIG_FILE_NAME in os.listdir():
        logging.info(f"{CONFIG_FILE_NAME} detected.")
        return 1
    elif os.getenv("TGCF_CONFIG"):
        logging.info(f"env var {CONFIG_ENV_VAR_NAME} detected.")
        if not ".env" in os.listdir():
            return 2
        else:
            logging.warning(
                f"If you can create files in your system, you should use tgcf.config.yml and not .env to define configuration. {tutorial_link}"
            )
            sys.exit(1)
    else:
        logging.warning(f"Configuration not found! {tutorial_link}")
        sys.exit(1)


CONFIG_TYPE = detect_config_type()


def read_config_file():
    if CONFIG_TYPE == 1:
        with open(CONFIG_FILE_NAME) as file:
            config_dict = yaml.full_load(file)
    elif CONFIG_TYPE == 2:
        config_env_var = os.getenv("TGCF_CONFIG")
        config_dict = yaml.full_load(config_env_var)
    else:
        logging.warning("This should never happen!")

    try:
        config = Config(**config_dict)
    except Exception as err:
        print(err)
        sys.exit(1)
    else:
        logging.info(config)
        return config


def update_config_file(config: Config):
    if CONFIG_TYPE == 1:
        with open(CONFIG_FILE_NAME, "w") as file:
            yaml.dump(config.dict(), file)
    elif CONFIG_TYPE == 2:
        logging.warning("Could not update config! As env var is used")


def get_env_var(name: str, optional=False):
    var = os.getenv(name)

    while not var:
        if optional:
            break
        var = input(f"Enter {name}: ")
    return var


API_ID = get_env_var("API_ID")
API_HASH = get_env_var("API_HASH")
USERNAME = get_env_var("USERNAME", optional=True)
SESSION_STRING = get_env_var("SESSION_STRING", optional=True)

if SESSION_STRING:
    SESSION = StringSession(SESSION_STRING)
else:
    SESSION = "tgcf"

CONFIG = read_config_file()

logging.info("config.py got executed")



import logging

from telethon import TelegramClient, events

from tgcf.config import API_HASH, API_ID, CONFIG, SESSION
from tgcf.plugins import apply_plugins
from tgcf.utils import send_message

from_to = {}

KEEP_LAST_MANY = 10000

existing_hashes = []

_stored = {}


for forward in CONFIG.forwards:
    from_to[forward.source] = forward.dest


class EventUid:
    def __init__(self, event):
        self.chat_id = event.chat_id
        try:
            self.msg_id = event.id
        except:
            self.msg_id = event.deleted_id

    def __str__(self):
        return f"chat={self.chat_id} msg={self.msg_id}"

    def __eq__(self, other):
        return self.chat_id == other.chat_id and self.msg_id == other.msg_id

    def __hash__(self) -> int:
        return hash(self.__str__())


async def new_message_handler(event):
    chat_id = event.chat_id

    if chat_id not in from_to:
        return
    logging.info(f"New message received in {chat_id}")
    message = event.message

    global _stored

    event_uid = EventUid(event)

    length = len(_stored)
    exceeding = length - KEEP_LAST_MANY

    if exceeding > 0:
        for key in _stored:
            del _stored[key]
            break

    to_send_to = from_to.get(chat_id)

    if to_send_to:
        if event_uid not in _stored:
            _stored[event_uid] = []

        message = apply_plugins(message)
        if not message:
            return
        for recipient in to_send_to:
            fwded_msg = await send_message(event.client, recipient, message)
            _stored[event_uid].append(fwded_msg)

    existing_hashes.append(hash(message.text))


async def edited_message_handler(event):
    message = event.message

    chat_id = event.chat_id

    if chat_id not in from_to:
        return

    logging.info(f"Message edited in {chat_id}")

    event_uid = EventUid(event)

    message = apply_plugins(message)

    if not message:
        return

    fwded_msgs = _stored.get(event_uid)

    if fwded_msgs:
        for msg in fwded_msgs:
            await msg.edit(message.text)
        return

    to_send_to = from_to.get(event.chat_id)

    for recipient in to_send_to:
        await send_message(event.client, recipient, message)


async def deleted_message_handler(event):
    chat_id = event.chat_id
    if chat_id not in from_to:
        return

    logging.info(f"Message deleted in {chat_id}")

    event_uid = EventUid(event)
    fwded_msgs = _stored.get(event_uid)
    if fwded_msgs:
        for msg in fwded_msgs:
            await msg.delete()
        return


ALL_EVENTS = {
    "new": (new_message_handler, events.NewMessage()),
    "edited": (edited_message_handler, events.MessageEdited()),
    "deleted": (deleted_message_handler, events.MessageDeleted()),
}


def start_sync():
    client = TelegramClient(SESSION, API_ID, API_HASH)
    for key, val in ALL_EVENTS.items():
        logging.info(f"Added event handler for {key}")
        client.add_event_handler(*val)
    client.start()
    client.run_until_disconnected()
